#! /usr/bin/python
# this is the first comment

import sys, os, optparse, numpy

class GenPovFile():

    def __init__(self, file_in, file_out, png_file):
        self.grid_points     = None
        self.scalar_amplify  = 1000.0 #This value is optimized for 128^2 grids, but will be rescaled for any size of grid.
        self.min_y           = 1
        self.max_y           = -1
        self.max_x           = 0
        self.max_z           = 0
        self.color_map       = \
"""[0 color Blue]
[1.0 color White]"""
        self.scale           = 1.0
        self.scale_y         = 1.0
        self.Coords          = None
        self.file_in         = file_in
        self.povrayfile      = file_out
        self.png_file        = png_file
        #self.color_map       = self.nm_color_map

    def readData(self):
        LL = [s.strip() for s in open(self.file_in, "r").readlines() if len(s.strip())>0]
        self.grid_points = int(numpy.sqrt(len(LL))) #Assume square lattice
        self.scalar_amplify  = self.scalar_amplify*(self.grid_points/128.0)
        self.Coords          = numpy.zeros((self.grid_points, self.grid_points), dtype=float)
        for line in LL:
            s_s = [q.strip() for q in line.split(" ")]
            (i, j, re_part, im_part) = (int(s_s[0]), int(s_s[1]), float(s_s[2]), float(s_s[3]))
            self.Coords[i, j] = numpy.sqrt(numpy.power(re_part, 2) + numpy.power(im_part, 2))

    def findMinMaxZ(self):
        self.min_y = self.Coords[0, 0]
        self.max_y = self.Coords[0, 0]
        for i in range(self.grid_points):
            for j in range(self.grid_points):
                if self.Coords[i, j] < self.min_y:
                    self.min_y = self.Coords[i, j]
                if self.max_y < self.Coords[i, j]:
                    self.max_y = self.Coords[i, j]

        self.scale_y = 1.0/(self.max_y - self.min_y)
        self.max_x = self.grid_points
        self.max_z = self.grid_points
        
    def drawTriangles(self):        
        outstring = """#include \"colors.inc\"
#include \"textures.inc\"
#include \"glass.inc\"
background{Black}"""

        outstring += """camera { angle 10
location <%s, %s, %s>
look_at <%s, %s, %s>}\n""" % (self.scalar_amplify, self.scalar_amplify/2, self.scalar_amplify/2, 0, 0, 0)
        
        outstring += """light_source { <%s, %s, %s> color White shadowless}
#declare surfface = union{ \n""" % (0, self.scalar_amplify, 0)
        
        for i in range(self.grid_points-1):
            for j in range(self.grid_points-1):

                outstring += 'triangle{'
                outstring += '<%s, %s, %s>' % (str(i - self.max_x/2.0), 
                                               str(self.scale_y*(self.Coords[i, j] - self.min_y)),
                                               str(j - self.max_z/2.0))
                outstring += '<%s, %s, %s>' % (str(i + 1 - self.max_x/2.0),
                                               str(self.scale_y*(self.Coords[i + 1, j] - self.min_y)), 
                                               str(j - self.max_z/2.0))
                outstring += '<%s, %s, %s>\n' % (str(i - self.max_x/2.0), 
                                                 str(self.scale_y*(self.Coords[i, j + 1] - self.min_y)),
                                                 str(j + 1 - self.max_z/2.0))
                outstring += """pigment{gradient y
color_map {
%s
}}}\n""" % self.color_map
                
                outstring += 'triangle{'
                outstring += '<%s, %s, %s>' % (str(i + 1 - self.max_x/2.0), 
                                               str(self.scale_y*(self.Coords[i + 1, j] - self.min_y)), 
                                               str(j - self.max_z/2.0))
                outstring += '<%s, %s, %s>' % (str(i - self.max_x/2.0),
                                               str(self.scale_y*(self.Coords[i, j + 1] - self.min_y)),
                                               str(j + 1 - self.max_z/2.0))
                outstring += '<%s, %s, %s>\n' % (str(i + 1 - self.max_x/2.0),
                                                 str(self.scale_y*(self.Coords[i + 1, j + 1] - self.min_y)),
                                                 str(j + 1 - self.max_z/2.0))
                outstring += """pigment{gradient y
color_map {
%s
}}}\n""" % self.color_map
        outstring +='}\n'
        outstring += 'object{surfface scale<1.33, 10, 1>}\n'
        f  = open(self.povrayfile , 'w')
        f.write(outstring)
        f.close()

    def execute(self):
        self.readData()
        self.findMinMaxZ()
        self.drawTriangles()
        os.system('povray %s +O./%s' % (self.povrayfile, self.png_file))

def main():
    parser = optparse.OptionParser()
    parser.add_option('--in_file', metavar="SIZE", action="store", dest="in_file", help="Specify the input file", default = None, type = str)
    parser.add_option('--pov_file', metavar="POVFILE", action="store", dest="pov_file", help="File to be fed to POV-RAY for rendering", default="surface.pov", type = str)
    parser.add_option('--png_file', metavar="PNGFILE", action="store", dest="png_file", help="Final output PNG file", default="surface.png", type = str)
    
    (args, remainder) = parser.parse_args()    

    Y = GenPovFile(args.in_file, args.pov_file, args.png_file)
    Y.execute()

if __name__ == "__main__":
    main()
