#include <iostream>
#include <fstream>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include "Constants.cpp"

unsigned long int step = 0;

typedef struct {
  long double phi[N_x][N_y];
  long double vel[N_x][N_y]; /* Phi is complex and we track the 
			    real and imaginary parts separately. */
} Coords; 

Coords* const T1 = new Coords;
Coords* const T2 = new Coords;

inline void periodic_boundaries(int i, int j, int& corr_i, int& corr_j){
  (i < 0) ? (corr_i = N_x+i):((i > N_x-1) ? (corr_i = i - N_x):corr_i = i);
  (j < 0) ? (corr_j = N_y+j):((j > N_y-1) ? (corr_j = j - N_y):corr_j = j);
}

inline void reset_coords(Coords& curr_pos){
  for(int i = 0; i < N_x; ++i){
    for(int j = 0; j < N_y; ++j){
      curr_pos.phi[i][j] = 0;
      curr_pos.vel[i][j] = 0;
    }
  }
}

inline void sync_coords(Coords& new_pos, Coords& old_pos){ 
  for(int i = 0; i < N_x; ++i){
    for(int j = 0; j < N_y; ++j){
	new_pos.phi[i][j] = old_pos.phi[i][j];
	new_pos.vel[i][j] = old_pos.vel[i][j];
    }
  }
}

inline void prefetch_data(int i, int j, Coords& Future_pos, Coords& Curr_pos){
  const int stride = 1;
  int i_plus, i_minus, i_shift;
  int j_plus, j_minus, j_shift;

  int ci_plus, ci_minus, ci_shift;
  int cj_plus, cj_minus, cj_shift;


  i_shift = i;
  j_shift = j + stride;
  i_plus  = i_shift + 1;
  i_minus = i_shift - 1;
  j_plus  = j_shift + 1;
  j_minus = j_shift - 1;

  periodic_boundaries(i_shift, j_shift, ci_shift, cj_shift);
  periodic_boundaries(i_plus,  j_plus , ci_plus , cj_plus );
  periodic_boundaries(i_minus, j_minus, ci_minus, cj_minus);

  if(j_shift < N_y){
    __builtin_prefetch(&Future_pos.phi[ci_shift][cj_shift], 1, 1);
    __builtin_prefetch(&Future_pos.vel[ci_shift][cj_shift], 1, 1);
    
    __builtin_prefetch(&Curr_pos.phi[ci_shift][cj_shift], 1, 1);
    __builtin_prefetch(&Curr_pos.vel[ci_shift][cj_shift], 1, 1);
    
    __builtin_prefetch(&Curr_pos.phi[ci_shift][cj_plus], 1, 1);
    __builtin_prefetch(&Curr_pos.vel[ci_shift][cj_plus], 1, 1);
    
    __builtin_prefetch(&Curr_pos.phi[ci_shift][cj_minus], 1, 1);
    __builtin_prefetch(&Curr_pos.vel[ci_shift][cj_minus], 1, 1);
    
    __builtin_prefetch(&Curr_pos.phi[ci_plus][cj_shift], 1, 1);
    __builtin_prefetch(&Curr_pos.vel[ci_plus][cj_shift], 1, 1);
    
    __builtin_prefetch(&Curr_pos.phi[ci_minus][cj_shift], 1, 1);
    __builtin_prefetch(&Curr_pos.vel[ci_minus][cj_shift], 1, 1);
  }
}


inline void read_data(Coords& temp_pos){
  int junk_i, junk_j, temp;
  const char* infile = std::string("initial.dat").c_str();
  std::ifstream fin;
  fin.open(infile);
  if (fin.is_open()){
    for(int i = 0; i < N_x; ++i){
      for(int j = 0; j < N_y; ++j){
	fin >> junk_i;
	fin >> junk_j;
	fin >> temp_pos.phi[i][j];
	fin >> temp; //temp_pos.phi[1][i][j];
      }
    }
  }
  else{
    std::cout << " Error: could not open " << infile << std::endl;
    exit(-1);
  }
  fin.close();
}

/******************************************************************************/
inline long double ScalarRealPart(Coords& Curr_pos, int i, int j){

  int c_i_P, c_j_P; /* m + j "m Plus  j"*/
  int c_i_M, c_j_M; /* m - j "m Minus j"*/
  long double tmp0;

  tmp0 = 0;
  
  periodic_boundaries(i+1, j, c_i_P, c_j_P);      
  periodic_boundaries(i-1, j, c_i_M, c_j_M);
  
  tmp0 = tmp0 + 
    pow(1/a, 2)*(Curr_pos.phi[c_i_P][c_j_P] - 
		 2*Curr_pos.phi[i][j] + 
		 Curr_pos.phi[c_i_M][c_j_M]);

  periodic_boundaries(i, j+1, c_i_P, c_j_P);      
  periodic_boundaries(i, j-1, c_i_M, c_j_M);
  
  tmp0 = tmp0 + 
    pow(1/a, 2)*(Curr_pos.phi[c_i_P][c_j_P] - 
		 2*Curr_pos.phi[i][j] + 
		 Curr_pos.phi[c_i_M][c_j_M]);
  
  return(tension*tmp0);
}

/******************************************************************************/
inline void UpdateRealScalar(Coords& Future_pos, Coords& Curr_pos, int i, int j){
  Future_pos.phi[i][j] = Curr_pos.phi[i][j] + dt*Future_pos.vel[i][j]; 
}

/******************************************************************************/
inline void UpdateVelRealScalar(Coords& Future_pos, Coords& Curr_pos, int i, int j){
  Future_pos.vel[i][j] = Curr_pos.vel[i][j] + 
    dt*(ScalarRealPart(Curr_pos, i, j) - H*Curr_pos.vel[i][j]);
}

/******************************************************************************/
inline void Evolve(Coords& Future_pos, Coords& Curr_pos){
  
  for(int i = 0; i < N_x; ++i){
    for(int j = 0; j < N_y; ++j){
      prefetch_data(i, j+1, Future_pos, Curr_pos);
      UpdateVelRealScalar(Future_pos, Curr_pos, i, j);
    }
  }
  for(int i = 0; i < N_x; ++i){
    for(int j = 0; j < N_y; ++j){
      prefetch_data(i, j+1, Future_pos, Curr_pos);
      UpdateRealScalar(Future_pos, Curr_pos, i, j);
    }
  }
}

void writeScalars(unsigned int local_frame){
  char buffer[16];
  sprintf(buffer, "%03d", local_frame);
  std::string curr_scalar_out("frame" + std::string(buffer) + ".txt");
  std::ofstream scalar_out(curr_scalar_out.c_str());
  scalar_out.precision(4);
  for(int ix = 0; ix < N_x; ++ix){
    for(int iy = 0; iy < N_y; ++iy){
      if((ix%grid_reduction == 0)&&(iy%grid_reduction == 0)){
	scalar_out << ix << " " << iy << " "
		   << T1->phi[ix][iy] 
		   << " " 
		   << 0 
		   << " "  << std::endl;  
	
      }
    }
  }
  scalar_out.close();
}

int main(int argc, char* argv[]){  
  char num_buff[16];
  reset_coords(*T1);
  reset_coords(*T2); 

  read_data(*T1);
  sync_coords(*T2, *T1);
    
  while(step < Steps){//Start the actual evolution
    ++step;
    
    Evolve(*T2, *T1);
    if(step%reduction == 0){
      sprintf(num_buff, "%03d", static_cast<int>(step/reduction));
      std::cout << "Frame " + std::string(num_buff) + "/" 
		<< frames_total << " done" << std::endl;
      writeScalars(static_cast<int>(step/reduction));
    }
    sync_coords(*T1, *T2);
  }
  delete T1;
  delete T2;
  return(0);
}

  
